# -*- coding: utf-8 -*-
#
# Expyriment documentation build configuration file, created by
# sphinx-quickstart on Fri Apr  5 15:08:26 2013.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys, os, time

sys.path.append(os.path.abspath(os.path.join("../../setup.py", "..")))
from setup import get_version_info_from_file, \
    get_version_info_from_release_info, \
    get_version_info_from_git
sys.path.remove(os.path.abspath(os.path.join("../../setup.py", "..")))


# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath('.'))

# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.autosummary',
              'sphinx.ext.doctest',
              #'sphinx.ext.viewcode',
              'numpydoc']

autosummary_generate = True
numpydoc_show_class_members = False

# How to install numpydoc? --> sudo easy_install numpydoc

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'Expyriment'
copyright = u'{0}, Florian Krause & Oliver Lindemann'.format(
    time.localtime().tm_year)

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#

# Check if we are building/installing from a built archive/distribution
filename = os.path.abspath(os.path.join(os.path.split(__file__)[0], '..',
                                        '..', 'expyriment', '__init__.py'))
version_info = get_version_info_from_file(filename)

# If not, we are building/installing from source
if version_info[0] == '':

    # Are we building/installing from a source archive/distribution?
    version_info = get_version_info_from_release_info()
    if version_info[2].startswith("$Format:"):

        # Are we building/installing from the GitHub repository?
        try:
            from subprocess import Popen, PIPE
            proc = Popen(['git', 'rev-list', '--max-parents=0', 'HEAD'],
                         stdout=PIPE, stderr=PIPE)
            initial_revision = proc.stdout.readline()
            if not b'e21fa0b4c78d832f40cf1be1d725bebb2d1d8f10' in \
                                                            initial_revision:
                raise Exception
            version_info = get_version_info_from_git()
        except:
          raise RuntimeError("Building documentation failed!")

# The release name
release = version_info[0]

# The normal X.Y.Z version.
version_nr = release.split("-")[0]  #release[:5]

# The short X.Y version.
version = ".".join(version_nr.split(".")[:2])  #version_nr[:3]

rst_epilog = """
.. |expyriment-wheel| replace:: expyriment-{0}-py2.py3-none-any.whl
.. |expyriment-wheel-code| replace:: ``expyriment-{0}-py2.py3-none-any.whl``
""".format(release)


# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# The reST default role (used for this markup: `text`) to use for all documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []


# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'sphinx_rtd_theme' # 'default', 'sphinxdoc',  'haiku' 

# Temporary work-around for spacing problem between parameter and parameter
# type in the doc, see https://github.com/numpy/numpydoc/issues/215. The bug
# has been fixed in sphinx (https://github.com/sphinx-doc/sphinx/pull/5976) but
# through a change in sphinx basic.css except rtd_theme does not use basic.css.
# In an ideal world, this would get fixed in this PR:
# https://github.com/readthedocs/sphinx_rtd_theme/pull/747/files
def setup(app):
    app.add_stylesheet("basic.css")

html4_writer = True

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#html_theme_options = {
#    "sidebarbgcolor": "#222222",
#    "relbarbgcolor": "#222222",
#    "linkcolor": "#ff9632",
#    "visitedlinkcolor": "#ff9632",
#    "sidebarlinkcolor": "#ff9632",
#    "relbarlinkcolor": "#a046fa",
#    "sidebartextcolor": "#969696",
#    "relbartextcolor": "#a046fa",
#    "footerbgcolor": "#222222",
#    "footertextcolor": "#969696",
#    "headbgcolor": "#dddddd",
#    "headtextcolor": "black",
#
#}

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "xpy_button.png"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = 'favicon.ico'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []  #['_static']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
html_sidebars = {'**': ['globaltoc.html', 'searchbox.html']}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
html_domain_indices = ['py-modindex']

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'Expyrimentdoc'


# -- Options for LaTeX output --------------------------------------------------

# The paper size ('letter' or 'a4').
#latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
#latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
  ('index', 'Expyriment.tex', u'Expyriment Documentation',
   u'Florian Krause \\& Oliver Lindemann', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = False

# If true, show page references after internal links.
#latex_show_pagerefs = False

# If true, show URL addresses after external links.
#latex_show_urls = False

# Additional stuff for the LaTeX preamble.
#latex_preamble = ''

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_domain_indices = True


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'expyriment', u'Expyriment Documentation',
     [u'Florian Krause & Oliver Lindemann'], 1)
]
