"""The expyriment randomize module.

This module contains various functions for randomizing data

"""

__author__ = 'Florian Krause <florian@expyriment.org>,\
              Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''


from ._randomize import rand_int_sequence, rand_int, rand_element, rand_norm
from ._randomize import coin_flip, shuffle_list, make_multiplied_shuffled_list
